# TODO: https://github.com/changhsinlee/pytest-mock-examples/blob/master/tests/test_mocking_constant_mocker.py


# import asyncio
# from omega_tx.mock import Hygrometer
# import sys
#
#
# async def read():
#     async with Hygrometer() as tx:
#         sys.stdout.write(str(await tx.get()))
#
# asyncio.run(read())


# import asyncio
# from omega_tx import Barometer
#
#
# async def read_once():
#     async with Barometer('192.168.1.171') as tx:
#         print(await tx.get())
#
# asyncio.run(read_once())



"""Partial mock of iBTHX.

The purpose of this `mock` is to perform interface testing with more realistic sensor values.
"""

import asyncio
import random
import time

from omega_tx.driver import COMMANDS


# average and standard deviation of readings for a little realism
units = {
    '°C': (25.0, 5.0),
    '°F': (77, 10.0),
    'mbar/hPa': (1000.0, 50.0),
    'inHg': (30.0, 2.0),
    'mmHg': (748.0, 50.0),
    '%': (36.0, 20.0)
}


class Barometer:
    """Mock driver for iBTHX Omega transmitters.

    Records barometric pressure, ambient temperature, and humidity.
    """
    def __init__(self, *args, **kwargs):
        """Initialize the mock device."""
        super().__init__(*args, **kwargs)
        self.data = {}

        # this event loop "perturbs" readings for more realistic mocking
        loop = asyncio.get_event_loop()
        self.task = loop.create_task(self._perturb())

    async def __aenter__(self):
        """Support `async with` by entering a client session."""
        return self

    async def __aexit__(self, exception_type, exception_value, traceback):
        """Support `async with` by exiting a client session."""
        pass

    async def get(self):
        """Mocked reading from the transmitter."""
        await asyncio.sleep(0.1)  # more realistic time delay for writing/reading
        return self.data

    async def _perturb(self):
        """Make the values dance!"""
        self.data = {'Time in ms': int(time.time() * 1000)}
        while True:
            for command, desc in COMMANDS.items():
                self.data[desc] = round(random.gauss(*units.get(desc.split()[-1])), 1)
            await asyncio.sleep(1.0)


class Hygrometer:
    """Mock driver for iBTHX Omega transmitters.

    Records barometric pressure, ambient temperature, and humidity.
    """
    def __init__(self, *args, **kwargs):
        """Initialize the mock device."""
        super().__init__(*args, **kwargs)
        self.data = {}

        # this event loop "perturbs" readings for more realistic mocking
        loop = asyncio.get_event_loop()
        self.task = loop.create_task(self._perturb())

    async def __aenter__(self):
        """Support `async with` by entering a client session."""
        return self

    async def __aexit__(self, exception_type, exception_value, traceback):
        """Support `async with` by exiting a client session."""
        pass

    async def get(self):
        """Mocked reading from the transmitter."""
        await asyncio.sleep(0.1)  # more realistic time delay for writing/reading
        return self.data

    async def _perturb(self):
        """Make the values dance!"""
        self.data = {'Time in ms': int(time.time() * 1000)}
        readings = ['Temperature in °C', 'Relative Humidity in %', 'Dewpoint in °C']
        while True:
            for desc in readings:
                self.data[desc] = round(random.gauss(*units.get(desc.split()[-1])), 1)
            await asyncio.sleep(1.0)



"""Python driver for Omega transmitters.

Distributed under the GNU General Public License v2
Copyright (C) 2020 NuMat Technologies
"""
import asyncio
import logging
import sys

import aiohttp


# for the iBTHX-W transmitter
COMMANDS = {
    'SRTC': ('Temperature in °C', 'metric'),
    'SRTF': ('Temperature in °F', 'imperial'),
    'SRHb': ('Pressure in mbar/hPa', 'metric'),
    'SRHi': ('Pressure in inHg', 'imperial'),
    'SRHm': ('Pressure in mmHg', 'metric'),
    'SRH2': ('Relative Humidity in %', 'universal'),
    'SRDF2': ('Dewpoint in °F', 'imperial'),
    'SRDC2': ('Dewpoint in °C', 'metric')}

logger = logging.getLogger(__name__)


class Barometer:
    """Driver for the iBTHX-W Omega transmitter.

    Reads barometric pressure, ambient temperature, and relative humidity.
    """
    def __init__(self, address: str, port: str = 2000, timeout: float = 2.0,
                 unit_system = 'metric'):
        """Initialize the device for the iBTHX-W.

        Note that this constructor does not connect. Connection happens on call:
        `tx = await Barometer().connect()` or `async with Barometer() as tx`.

        Parameters
        ----------
        address : str
            Assigned iBTHX IP address.
        port : str
            Default port is 2000.
        timeout : float
            Applied both for establishing the connection as well as reading.
        unit_system : str
            Select either metric, imperial, or all units for the sensor request.

        Methods
        -------
        get()
            Bad reads are reported as None; an empty dictionary is returned if any write/read
            fails.
        """
        self.address = address
        self.port = port
        self.reader = None
        self.writer = None
        self.timeout = timeout
        self.data = None
        self.unit_system = unit_system

    async def __aenter__(self):
        """Support `async with` by entering a client session."""
        try:
            await self.connect()
        except Exception as err:  # noqa
            await self.__aexit__(*sys.exc_info())
            logger.error(f'Connection failed at address {self.address} and port {self.port}.')
            raise err
        return self

    async def __aexit__(self, exception_type, exception_value, traceback):
        """Support `async with` by exiting a client session."""
        if self.writer is not None:
            await self.disconnect()

    async def connect(self):
        """Establish the TCP connection with asyncio.streams.

        Refer to https://docs.python.org/3/library/asyncio-stream.html.
        """
        try:
            self.reader, self.writer = await asyncio.wait_for(
                asyncio.open_connection(self.address, self.port), timeout=self.timeout)
        except ConnectionRefusedError:
            logger.error('Failed connection attempt.')

    async def disconnect(self):
        """Close the underlying socket connection, if exists."""
        if self.writer is not None:
            self.writer.close()
            self.writer = None

    async def get(self):
        """Write and read from the IBTHX sensor.

        This method should not be used for time sensitive operations. However, this sensor
        is designed for `low resolution` (> 5 minute) monitoring.

        A connection is made when the user calls this method, even when the connect() method has
        not yet been explicitly called. Users can unplug (replug) sensors without the
        logging/control script being aware. This removes the need to handle re-connects outside the
        driver but the driver user is warned that the sensor has potentially been unplugged or lost
         power somewhere along the way.
        """
        if self.writer is None:
            logger.warning('TCP connection not created before the request.')
            await self.connect()
            logger.warning('A stream writer has been defined without the user explicitly calling '
                           'connect() method. Sensor may have been unplugged at some point.')

        self.data, response = {}, None
        commands = {
            command: (desc, unit) for command, (desc, unit) in COMMANDS.items()
            if unit == self.unit_system or unit == 'universal' or self.unit_system == 'all'}
        for command, (desc, units) in commands.items():
            self.data[desc] = None  # bad read value
            self.writer.write(f'*{command}\r'.encode())
            await self.writer.drain()

            try:
                fut = await asyncio.wait_for(self.reader.read(1024), timeout=self.timeout)
                response = fut.decode()
            except asyncio.TimeoutError:
                logger.warning(f'Failed to read on command {command} response, based on timeout '
                               f'of {self.timeout} s.')
                response = 'TIMEOUT!'

            try:
                if str(response) == 'ERROR!\r':  # response from IBTHX on malformed command
                    logger.error(f'Failed read from device; exited with {response} due to '
                                 f'malformed write command.')
                else:
                    self.data[desc] = float(response)
            except (ValueError, TypeError):
                logger.warning(f'Failed read from device on command {command}; unidentified'
                               f' error with response: {response}.')
        return self.data


class Hygrometer:
    """Driver for the iTHX-W Omega transmitter.

    Reads ambient temperature and relative humidity.
    """
    def __init__(self, address: str, timeout: float = 2.0):
        """Initialize the device for the iTHX-W.

        Parameters
        ----------
        address : str
            Assigned iBTHX IP address.
        timeout : float
            Applied both for establishing the connection as well as reading.

        Methods
        -------
        get()
            Bad reads are reported as -888.88.
        """
        self.address = address
        self.data = None
        self.session = None
        self.timeout = timeout

    async def __aenter__(self):
        """Support `async with` by entering a client session."""
        try:
            await self.connect()
        except Exception as err:  # noqa
            await self.__aexit__(*sys.exc_info())
        return self

    async def __aexit__(self, exception_type, exception_value, traceback):
        """Support `async with` by exiting a client session."""
        if self.session is not None:
            await self.disconnect()

    async def connect(self):
        """Establish a connector instance for making HTTP requests."""
        self.session = aiohttp.ClientSession()

    async def disconnect(self):
        """Close the connector instance used for making HTTP requests."""
        if self.session is not None:
            await self.session.close()

    async def get(self):
        """Read and parse hygrometer (iTHX-W) data from the hosted HTML page."""
        self.data, response, text = {}, None, None
        try:
            response = await asyncio.wait_for(
                self.session.get(f'http://{self.address}/postReadHtml?a='), self.timeout)  # noqa
            assert response.status == 200
            text = await response.text()
        except aiohttp.ClientConnectorError:
            logger.error('Failed to establish HTTP connector instance.')
        except AssertionError:
            logger.error(f'Failed to read from transmitter HTML page: {response.status}')

        if text:
            try:
                temp, humid, dew = text.split('\n')[1:4]
                self.data = {
                    'Temperature in °C': float(temp.split()[2]),
                    'Relative Humidity in %': float(humid.split()[2]),
                    'Dewpoint in °C': float(dew.split()[2]),
                }
            except ValueError:
                logger.error('Failed to properly parse the HTML.')
        return self.data



"""Python driver for Omega iTHX and iBTHX transmitters.

Distributed under the GNU General Public License v2
Copyright (C) 2020 NuMat Technologies

"""
import argparse
import json
import logging
import sys

import asyncio

from omega_tx.driver import Barometer, Hygrometer


log = logging.getLogger(__name__)

bash_red, color_off = '\033[0;31m', '\033[0m'


def _parse_args() -> argparse.Namespace:
    """Organize the command line arguments."""
    parser = argparse.ArgumentParser(description="Read Omega transmitters.")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="Verbose debug output")
    parser.add_argument(
        'model', help="Model of Omega transmitter. ibthx and ithx supported.",
        choices=['ibthx', 'ithx'])
    parser.add_argument(
        'address', help="The IP address of the Omega transmitter.")
    parser.add_argument(
        '-p', '--port', help="The port of the Omega transmitter (default 2000).", default=2000)
    parser.add_argument(
        '-t', '--timeout', help="Request timeout (default 2.0 s).", default=2.0, type=float)
    parser.add_argument(
        '-u', '--unit_system', help="Request the type of units to return (default metric).",
        default='metric', choices=['all', 'imperial', 'metric'])
    return parser.parse_args()


def command_line() -> int:
    """Command line tool for Omega transmitters."""
    args = _parse_args()
    logging.basicConfig(
        format="[%(asctime)s] %(levelname)s: %(message)s (%(filename)s:%(lineno)d)",
        level=logging.DEBUG if args.debug else logging.INFO)

    async def read_once():
        """Perform a single read from the transmitter."""
        try:
            async with Barometer(args.address, args.port,
                                 args.timeout, args.unit_system) if args.model == 'ibthx' \
                    else Hygrometer(args.address, args.timeout) as tx:
                sys.stdout.write(
                    json.dumps(await tx.get(), indent=4, sort_keys=True, ensure_ascii=False))
        except asyncio.TimeoutError:
            sys.stderr.write(f'{bash_red}Could not connect to device.{color_off}\n')

    asyncio.run(read_once())


if __name__ == '__main__':
    command_line()



"""Package manager setup for the omega transmitter driver."""
from setuptools import setup

with open('README.md', 'r') as in_file:
    long_description = in_file.read()

setup(
    name='omega-tx',
    version='0.1.0',
    description='Python driver for Omega iTHX-W and iBTHX-W transmitters.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/numat/omega-tx/',
    author='Ross Verploegh, PhD',
    author_email='ross@numat-tech.com',
    packages=['omega_tx'],
    install_requires=[
        'aiohttp>=3.7.2'
    ],
    entry_points={
        'console_scripts': [
            'omega-tx = omega_tx:command_line',
        ],
    },
    license='GPLv2',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Development Status :: 2 - Pre-Alpha',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces'
    ]
)
